﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ThinkGeo.MapSuite.Core;
using ThinkGeo.MapSuite.DesktopEdition;
using GisSharpBlog.NetTopologySuite.Index.Quadtree;
using GisSharpBlog.NetTopologySuite.Geometries;
using System.Collections.Specialized;
using System.Collections;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Windows.Threading;

namespace WpfApplication1
{
    #region PointsFeatureSource
    class PointsFeatureSource : FeatureSource
    {
 
        #region MemberTypes

        public class PointsFeatureCollection : KeyedCollection<String, Feature>
        {
            public PointsFeatureCollection() : base() { }

            protected override string GetKeyForItem(Feature f)
            {
                return f.Id;
            }
        }
        #endregion MemberTypes
        #region Fields

        private Collection<FeatureSourceColumn> EMPTY_COLUMNS = new Collection<FeatureSourceColumn>();
        private Collection<Feature> EMPTY_FEATURES = new Collection<Feature>();

        /* Coordinates to Feature ID */
        Quadtree featureTree;

        /* Feature ID to Feature lookups*/
        PointsFeatureCollection featureIdDictionary;

        /* Point ID to Feature ID lookups */
        Dictionary<Point, string> pointDictionary;

        /* Feature exclude list */
        PointsFeatureCollection featuresToExclude;

        /* Associated feature layer */
        PointsFeatureLayer layer;

        Window1 controller;

        private Object syncLock = new object();

        #endregion Fields
        #region Constructors
        public PointsFeatureSource(PointsFeatureLayer layer, Window1 controller)
            : base()
        {
            this.layer = layer;
            this.controller = controller;

            featureTree = new Quadtree();
            featureIdDictionary = new PointsFeatureCollection();
            pointDictionary = new Dictionary<Point, string>(15000);
            featuresToExclude = new PointsFeatureCollection();

            
            AddFeatures(controller.ReceiverPoints);
            AddFeatures(controller.LTUPoints);
            AddFeatures(controller.ShotPoints);
          
            //((INotifyCollectionChanged)controller.ReceiverPoints).CollectionChanged +=
            //    new NotifyCollectionChangedEventHandler(Points_CollectionChanged);
            //((INotifyCollectionChanged)controller.ShotPoints).CollectionChanged +=
            //    new NotifyCollectionChangedEventHandler(Points_CollectionChanged);
            //((INotifyCollectionChanged)controller.LTUPoints).CollectionChanged +=
            //    new NotifyCollectionChangedEventHandler(Points_CollectionChanged);

        }
        #endregion Constructors
        #region EventHandlers


        void Point_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {

            Point p = sender as Point;
            switch (e.PropertyName)
            {
                case "DisplayType":
                    controller.RefreshView(DispatcherPriority.Background);
                    break;
                default:
                    break; // Ignore

            }

        }

        void Points_CollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
        {

            switch (e.Action)
            {
                case NotifyCollectionChangedAction.Remove:
                    foreach (Point p in e.OldItems)
                    {
                        RemoveFeature(p);
                    }
                    break;
                case NotifyCollectionChangedAction.Add:
                    foreach (Point p in e.NewItems)
                    {
                        CreateFeature(p);

                    }
                    break;
                case NotifyCollectionChangedAction.Replace:
                    
                    foreach (Point p in e.OldItems)
                    {
                        RemoveFeature(p);
                    }

                    foreach (Point p in e.NewItems)
                    {
                        CreateFeature(p);
                    }
                    
                    break;
                case NotifyCollectionChangedAction.Reset:
               
                    if (e.NewItems != null)
                    {
                        foreach (Point p in e.NewItems)
                        {
                            Feature? nf = FindFeature(p);
                            if (nf != null)
                            {
                                RemoveFeature(p);
                                CreateFeature(p);
                            }
                            else
                            {
                                CreateFeature(p);
                            }
                        }

                    }
                    else
                    {
                        
                        featureTree = new Quadtree();
                        featureIdDictionary.Clear();
                        pointDictionary.Clear();


                    }
                    
                    break;
            }

            controller.RefreshView(DispatcherPriority.Background);
        }


        #endregion EventHandlers
        #region Properties
        protected override bool CanExecuteSqlQueryCore
        {
            get
            {
                return false;
            }
        }
        public override bool IsEditable
        {
            get
            {
                return false;
            }
        }
        public Collection<Feature> FeaturesToExclude
        {
            get { return featuresToExclude; }

        }
        #endregion Properties
        #region Methods
        private void AddFeatures(IList<Point> points)
        {

            foreach (Point p in points)
            {
                CreateFeature(p);
            }
        }
        private void UpdateFeature(Point p)
        {
            
            Feature? nf = FindFeature(p);
            if (nf == null)
            {
     
                return;
            }

            lock (this.syncLock)
            {
                RemoveFeature(p);
                CreateFeature(p);
            }

        }


        private void CreateFeature(Point p)
        {
   
            double lon = p.DisplayLon;
            double lat = p.DisplayLat;
            Feature f = new Feature(lon, lat);
            f.Tag = p;


            RectangleShape rectangle = f.GetBoundingBox();
            Envelope envelope = new Envelope(rectangle.UpperLeftPoint.X, rectangle.LowerRightPoint.X, rectangle.UpperLeftPoint.Y, rectangle.LowerRightPoint.Y);

            lock (this.syncLock)
            {
                featureTree.Insert(envelope, f.Id);

                // Feature ID to Feature lookups
                featureIdDictionary.Add(f);

                // Point to feature lookups
                pointDictionary.Add(p, f.Id);

            }

            p.PropertyChanged += new PropertyChangedEventHandler(Point_PropertyChanged);

        }

        private void RemoveFeature(Point p)
        {
  
            Feature? nf = FindFeature(p);
            if (nf == null)
            {
                return;
            }
            Feature f = nf.Value;
            RectangleShape rectangle = f.GetBoundingBox();
            Envelope envelope = new Envelope(rectangle.UpperLeftPoint.X, rectangle.LowerRightPoint.X, rectangle.UpperLeftPoint.Y, rectangle.LowerRightPoint.Y);
            lock (this.syncLock)
            {
                featureTree.Remove(envelope, f.Id);

                // Update dictionaries
                featureIdDictionary.Remove(f.Id);
                pointDictionary.Remove(p);
            }
            p.PropertyChanged -= new PropertyChangedEventHandler(Point_PropertyChanged);
        }

        public Feature? FindFeature(Point p)
        {
            lock (this.syncLock)
            {
                String fid = null;
                pointDictionary.TryGetValue(p, out fid);
                if (fid != null)
                    return featureIdDictionary[fid];
                return null;
            }
        }


        protected override Collection<Feature> GetAllFeaturesCore(IEnumerable<string> returningColumnNames)
        {

            Collection<Feature> result = new Collection<Feature>();
            lock (this.syncLock)
            {
                if (FeaturesToExclude.Count == 0)
                {
                    result = new Collection<Feature>(featureIdDictionary);
                }
                else
                {
                    foreach (Feature f in featureIdDictionary)
                    {
                        if (!this.featuresToExclude.Contains(f.Id))
                            result.Add(f);
                    }
                }

                return result;
            }
        }

        protected override Collection<Feature> GetFeaturesForDrawingCore(RectangleShape boundingBox, double screenWidth, double screenHeight, IEnumerable<string> returningColumnNames)
        {

            Collection<Feature> results = GetFeaturesBasedOnCachedBoundingBoxes(boundingBox);

            return results;
        }

        protected override Collection<Feature> GetFeaturesInsideBoundingBoxCore(RectangleShape boundingBox, IEnumerable<string> returningColumnNames)
        {

            Collection<Feature> results = GetFeaturesBasedOnCachedBoundingBoxes(boundingBox);

            return results;
        }

        protected override Collection<FeatureSourceColumn> GetColumnsCore()
        {
            return EMPTY_COLUMNS;
        }
        protected override int GetCountCore()
        {
            return featureIdDictionary.Count - FeatureIdsToExclude.Count;
        }

        protected override Collection<Feature> GetFeaturesByIdsCore(IEnumerable<string> ids, IEnumerable<string> returningColumnNames)
        {


            Collection<Feature> results = new Collection<Feature>();
            lock (this.syncLock)
            {
                foreach (string id in ids)
                {

                    if (featureIdDictionary.Contains(id))
                        results.Add(featureIdDictionary[id]);
                }

                return results;
            }
        }

        public Collection<Feature> GetFeaturesBasedOnCachedBoundingBoxes(RectangleShape extent)
        {


            Envelope envelope = new Envelope(extent.UpperLeftPoint.X, extent.LowerRightPoint.X, extent.UpperLeftPoint.Y, extent.LowerRightPoint.Y);
            lock (this.syncLock)
            {
                ArrayList RecordIds = (ArrayList)featureTree.Query(envelope);

                Collection<Feature> features = new PointsFeatureCollection();
                foreach (string key in RecordIds)
                {
                    if (!featuresToExclude.Contains(key) && featureIdDictionary.Contains(key))
                        features.Add((Feature)featureIdDictionary[key]);
                }

                return features;
            }
        }

        #endregion Methods


    }
    #endregion PointsFeatureSource

    #region PointsFeatureLayer
    class PointsFeatureLayer : FeatureLayer
    {
 
        #region Fields
        Window1 controller = null;
        LayoutPointStyle pointStyle = null;
        #endregion Fields
        #region Constructors
        public PointsFeatureLayer(Window1 controller)
        {
            this.controller = controller;
            FeatureSource = new PointsFeatureSource(this, controller);
            DrawingQuality = DrawingQuality.HighSpeed;
            pointStyle = new LayoutPointStyle();
            ZoomLevelSet.ZoomLevel01.CustomStyles.Add(pointStyle);
            ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;
        }
        public override bool HasBoundingBox
        {
            get
            {
                return true;
            }
        }
        #endregion Constructors
    }
    #endregion PointsFeatureLayer

}
